import json
import re
import hexdump

PRINT_INSTR_ARRAY = True
PRINT_OUTPUT_HEX = True

file_name_input = str(input("Input file: "))
source_file_name = f"programs/{file_name_input}.txt"

with open("reserved.json") as reserved_file:
    opcodes = json.load(reserved_file)["res"]

with open("opcodes.json") as json_opcodes:
    opcodes_values = json.load(json_opcodes)

with open(source_file_name, "r") as source_file:
    lines = source_file.readlines()

for i in range(len(lines)):
    lines[i] = lines[i].replace("\n", "")
    lines[i] = lines[i].split(";", 1)[0]

#lines = [line for line in lines if line]

#print(lines)

print(f"Assembling {source_file_name}...")

instructions = []
label_addresses = {}
static_variables = {}
expressions = {}
expressions_results = {}

def number_label_address(arg):
    if arg[0] == "#" or arg[0] == "$":
        if arg[1:3] == "0x": return int(arg[1:], 16)
        if arg[1:3] == "0b": return int(arg[1:], 2)
        return int(arg[1:])
    return arg


def process_ldm(cmd):
    try:
        if cmd[1][0] == "$" and cmd[1][1] in ["A", "B", "C", "D"]:
            instructions.append([1, f"ldm {cmd[1]},{cmd[2]}"])
        else:
            instructions.append([2, f"ldm <addr>,{cmd[2]}", number_label_address(cmd[1])])
    except:
        return True
    return False

def process_stm(cmd):
    try:
        if cmd[2][0] == "$" and cmd[2][1] in ["A", "B", "C", "D"]:
            instructions.append([1, f"stm {cmd[1]},{cmd[2]}"])
        else:
            instructions.append([2, f"stm {cmd[1]},<addr>", number_label_address(cmd[2])])
    except:
        return True
    return False

def process_mov(cmd):
    try:
        instructions.append([1, f"mov {cmd[1]},{cmd[2]}"])
    except:
        return True
    return False

def process_swp(cmd):
    try:
        mirror = False

        if cmd[1] == "B" and cmd[2] == "A": mirror = True
        if cmd[1] == "C" and cmd[2] == "A": mirror = True
        if cmd[1] == "D" and cmd[2] == "A": mirror = True
        if cmd[1] == "C" and cmd[2] == "B": mirror = True
        if cmd[1] == "D" and cmd[2] == "B": mirror = True
        if cmd[1] == "D" and cmd[2] == "C": mirror = True

        if not mirror:
            instructions.append([1, f"swp {cmd[1]},{cmd[2]}"])
        else:
            instructions.append([1, f"swp {cmd[2]},{cmd[1]}"])
    except:
        return True
    return False

def process_ldr(cmd):
    try:
        instructions.append([2, f"ldr {cmd[1]},<value>", number_label_address(cmd[2])])
    except:
        return True
    return False

def process_alu_operation(cmd):
    try:
        if cmd[1] not in ["A","B","C","D"]:
            if cmd[1][0] == "#": 
                instructions.append([2, f"{cmd[0]} <value>", number_label_address(cmd[1])])
            else:
                instructions.append([2, f"{cmd[0]} <addr>", number_label_address(cmd[1])])
        else:
            instructions.append([1, f"{cmd[0]} {cmd[1]}"])
    except:
        return True
    return False


def process_alu_single_byte_operation(cmd):
    try:
        if cmd[1] not in ["A","B","C","D"]:
            #return True
            instructions.append([2, f"{cmd[0]} <addr>", number_label_address(cmd[1])])
        else:
            instructions.append([1, f"{cmd[0]} {cmd[1]}"])
    except:
        return True
    return False

def process_stack_operations(cmd):
    try:
        if cmd[1] not in ["A","B","C","D"]:
            return True
        else:
            instructions.append([1, f"{cmd[0]} {cmd[1]}"])
    except:
        return True
    return False

def process_jumps_calls(cmd):
    try:
        instructions.append([2, f"{cmd[0]} <addr>", number_label_address(cmd[1])])
    except:
        return True
    return False

def process_single_arg_instr(cmd):
    try:
        instructions.append([1, f"{cmd[0]}"])
    except:
        return True
    return False


def handle_label_variable(cmd):
    value_size = 0
    value = 0

    if len(cmd) == 1:
        instructions.append([0, f"{cmd[0][:-1]}"])

    elif len(cmd) > 1:
        instructions.append([0, f"{cmd[0][:-1]}"])
        if cmd[1][0] == "#":
            if cmd[1][1:3] == "0x":
                value = int(cmd[1][1:], 16)
            elif cmd[1][1:3] == "0b":
                value = int(cmd[1][1:], 2)
            else:
                value = int(cmd[1][1:])

            value_size = (value.bit_length() + 7) // 8

        elif cmd[1] == "ascii":
            value_size = len(cmd[2])
            value = cmd[2]

        elif cmd[1] == "asciiz":
            value_size = len(cmd[2]) + 1
            value = f"{cmd[2]}\0"
        
        instructions.append([value_size, value])

    label_addresses[cmd[0][:-1]] = 0

    return False

def handle_static_variable(cmd):
    if len(cmd) != 3:
        return True

    if cmd[1] == "=":
        value = 0
        offset = 0
        if cmd[2][0] == "#": offset = 1
        if cmd[2][0+offset:2+offset] == "0x":
            value = int(cmd[2][0+offset:], 16)
        elif cmd[2][0+offset:2+offset] == "0b":
            value = int(cmd[2][0+offset:], 2)
        else:
            value = int(cmd[2][0+offset:])

        static_variables[cmd[0]] = value

    else:
        return True

    return False


def process_line(cmd, line_number):

    error = False

    if cmd[0] in opcodes:
        if cmd[0] == "ldm":
            error = process_ldm(cmd)
        if cmd[0] == "stm":
            error = process_stm(cmd)
        if cmd[0] == "mov":
            error = process_mov(cmd)
        if cmd[0] == "swp":
            error = process_swp(cmd)
        if cmd[0] == "ldr":
            error = process_ldr(cmd)
        if cmd[0] in ["add", "sub", "adc", "sbb", "and", "or", "nand", "nor", "xor", "xnor", "cmp"]:
            error = process_alu_operation(cmd)
        if cmd[0] in ["inc", "dec", "shl", "shr", "rol", "ror", "tmc", "inv"]:
            error = process_alu_single_byte_operation(cmd)
        if cmd[0] in ["push", "pop", "peek"]:
            error = process_stack_operations(cmd)
        if cmd[0] in ["jump", "jmc", "jmz", "jmn", "jmp", "jmnc", "jmnz", "jmnn", "jmnp",  "call", "cllc", "cllz", "clln", "cllp ", "cllnc", "cllnz", "cllnn", "cllpn"]:
            error = process_jumps_calls(cmd)
        if cmd[0] in ["ret", "rtc", "rtz", "rtn", "rtp", "rtnc", "rtnz", "rtnn", "rtnp", "clf", "hlt"]:
            error = process_single_arg_instr(cmd)

    else:
        #print(cmd)
        if cmd[0][-1] == ":":
            error = handle_label_variable(cmd)
        else:
            error = handle_static_variable(cmd)

    if error:
        print(f"Error in instruction {cmd} at line {line_number}!")

    return



for i in range(len(lines)):
    cmd = lines[i]

    if not '"' in cmd:
        cmd = cmd.replace(",", " ")
        cmd = cmd.split()
    else:
        string = re.findall(r'"([^"]*)"', cmd)[0]
        cmd = cmd.replace(f'"{string}"', "")
        cmd = cmd.replace(",", " ")
        cmd = cmd.split()
        cmd.append(string)

    #print(cmd)

    if cmd: process_line(cmd, i+1)

total_bytes = 0

for i in range(len(instructions)):
    if instructions[i][0] == 0:
        label_addresses[instructions[i][1]] = total_bytes
    else:
        total_bytes += instructions[i][0]

if total_bytes > 256:
    print("WARNING - PROGRAM LARGER THAN 256 BYTES!")

for i in range(len(instructions)):
    if instructions[i][0] != 0:
        try:
            label = instructions[i][2]
            if type(label) == str:
                if label in label_addresses.keys():
                    instructions[i][2] = label_addresses[label]
                elif label in static_variables.keys():
                    instructions[i][2] = static_variables[label]
                elif label in expressions.keys():
                    instructions[i][2] = expressions[label]
                else:
                    print(f"Error: Undefined label {label}!")
                    break
        except:
            pass


if PRINT_INSTR_ARRAY:
    print("Instruction array:")
    print(instructions)

out_file = open("output.bin", "wb")

for i in range(len(instructions)):
    if instructions[i][0] != 0:

        if instructions[i][1] in opcodes_values.keys():
            opcode = opcodes_values[instructions[i][1]]
            out_file.write(bytes.fromhex(opcode))

            if len(instructions[i]) == 3:
                out_file.write(bytes([instructions[i][2]]))

        else:
            value = instructions[i][1]

            if type(value) == int:
                out = value.to_bytes(instructions[i][0], "big")
            else:
                out = value.encode("ascii")

            out_file.write(out)

if total_bytes < 256:
    for i in range(256-total_bytes):
        out_file.write(bytes([0]))

out_file.close()


input_file = open("output.bin", mode="rb")
program = bytearray(input_file.read())

output = ""
for i in range(len(program)):
    if i != (len(program) - 1):
        output += "0x{:02x}, ".format(program[i])
    else:
        output += "0x{:02x}".format(program[i])

out_file = open("output_loader.txt", "w")
out_file.write(output)
out_file.close()

print("Done.\n")

if PRINT_OUTPUT_HEX:
    print("Hex:")
    hexdump.hexdump(program)

input()